class CommandBasedFrameworkError(Exception):
    """Base exception for all framework exceptions."""


class SchedulerExistsError(CommandBasedFrameworkError):
    """A scheduler instance already exists.

    Only one scheduler may exist at any given time. More than one would
    allow undefined behavior to occur in terms of resource allocation
    and usage.

    Instead, use :py:class:`~command_based_framework.scheduler.Scheduler`'s
    :py:obj:`~command_based_framework.scheduler.SchedulerMeta.instance`
    or delete all references to the current scheduler and recreate it.
    """  # noqa: E501



from command_based_framework.commands import Command


def test_name() -> None:
    """Verify the name of commands are set properly."""
    class MyCommand(Command):
        def is_finished(self) -> bool:
            return False
        def execute(self) -> None:
            return None

    command1 = MyCommand()
    command2 = MyCommand(name="HelloWorld")

    assert command1.name == "MyCommand"
    assert command2.name == "HelloWorld"
    assert str(command1) == "<MyCommand>"


def test_needs_interrupt() -> None:
    """Verify the needs interrupt property sets properly."""

    class MyCommand(Command):

        def handle_exception(self,*_) -> bool:
            return False

        def is_finished(self) -> bool:
            return True

        def execute(self) -> None:
            raise ValueError("test error")

    command = MyCommand()

    # Simulate executing a command
    with command:
        command.execute()

    # Verify the property is set
    assert command.needs_interrupt

    # Verify reading the property reset it
    assert not command.needs_interrupt



import pytest

from command_based_framework.commands import Command
from command_based_framework.scheduler import Scheduler
from command_based_framework.subsystems import Subsystem

def test_name() -> None:
    """Verify the name of subsystems are set properly."""
    scheduler = Scheduler.instance or Scheduler()

    class MySubsystem(Subsystem):
        def is_finished(self) -> bool:
            return False
        def execute(self) -> None:
            return None

    subsystem1 = MySubsystem()
    subsystem2 = MySubsystem(name="HelloWorld")

    assert subsystem1.name == "MySubsystem"
    assert subsystem2.name == "HelloWorld"


def test_current_and_default_commands() -> None:
    """Verify current and default commands get set"""
    scheduler = Scheduler.instance or Scheduler()

    class MyCommand(Command):
        def execute(self) -> None:
            return super().execute()

        def is_finished(self) -> bool:
            return super().is_finished()

    class MySubsystem(Subsystem):
        def periodic(self) -> None:
            return super().periodic()

    subsystem = MySubsystem()
    command = MyCommand(None, subsystem)
    command_no_requirement = MyCommand()

    # Verify the commands are set
    subsystem.default_command = command
    assert subsystem.default_command == command
    assert subsystem.current_command == None

    subsystem.current_command = command
    assert subsystem.current_command == command

    subsystem.current_command = None
    subsystem.default_command = None
    assert subsystem.default_command == None
    assert subsystem.current_command == None

    # Verify the subsystem rejects default commands that don't require
    # it
    with pytest.raises(ValueError):
        subsystem.default_command = command_no_requirement



from command_based_framework._common import ContextManagerMixin


def test_ctx_man_mixin() -> None:
    """Verify the ctx manager mixin works correctly."""

    class MyClass(ContextManagerMixin):
        called_handled_exception = False

        def handle_exception(self, *exc):
            self.called_handled_exception = True

        def raise_error(self) -> None:
            raise ValueError("test error")

    with MyClass() as myclass:
        myclass.raise_error()
        assert myclass.called_handled_exception

    with MyClass() as myclass:
        # Do nothing, just exit
        pass



import math
from threading import Thread
import time

import pytest

from command_based_framework.actions import Action, Condition
from command_based_framework.commands import Command, ParallelCommandGroup, SequentialCommandGroup
from command_based_framework.scheduler import Scheduler
from command_based_framework.subsystems import Subsystem


def test_tracking_scheduler_instances() -> None:
    """Verify only one scheduler instance exists at any one time."""
    from command_based_framework.exceptions import SchedulerExistsError

    # Verify no scheduler has been set
    assert Scheduler.instance == None

    # Create a scheduler and verify it is tracked
    s = Scheduler()
    assert Scheduler.instance == s

    # Ensure no new scheduler can be created
    with pytest.raises(SchedulerExistsError):
        Scheduler()

    # Delete the reference and ensure a new scheduler can be created
    del s
    t = Scheduler()

    # Once again, ensure no new scheduler can be created
    with pytest.raises(SchedulerExistsError):
        Scheduler()


def test_setting_clock_speed() -> None:
    """Verify the clock speed is set properly."""
    scheduler = Scheduler.instance or Scheduler()

    assert math.isclose(scheduler.clock_speed, 1 / 60)

    # Set the clock speed normally
    scheduler.clock_speed = 1 / 50

    # Ensure setting at or below 0 raises value errors
    with pytest.raises(ValueError):
        scheduler.clock_speed = 0

    with pytest.raises(ValueError):
        scheduler.clock_speed = -1

    # Ensure the clock speed is still at 1/50
    assert math.isclose(scheduler.clock_speed, 1 / 50)


def test_rebinding_same_command() -> None:
    """Verify actions are bound to commands correctly."""
    scheduler = Scheduler.instance or Scheduler()
    scheduler._reset_all_stacks()

    # Verify the stack is empty
    assert not scheduler._actions_stack

    # Add a new mapping
    class MyAction(Action):

        def poll(self) -> bool:
            return True

    class MyCommand(Command):

        def is_finished(self) -> bool:
            return True

        def execute(self) -> None:
            return None

    action = MyAction()
    command = MyCommand()
    condition = Condition.when_activated

    scheduler.bind_command(action, command, condition)
    assert action in scheduler._actions_stack
    assert scheduler._actions_stack[action] == {condition: {command}}

    # Rebind the command, just on a different condition
    new_condition = Condition.when_deactivated
    scheduler.bind_command(action, command, new_condition)
    assert action in scheduler._actions_stack
    assert len(scheduler._actions_stack) == 1
    assert len(scheduler._actions_stack[action]) == 2
    assert len(scheduler._actions_stack[action][condition]) == 0
    assert scheduler._actions_stack[action][new_condition] == {command}


def test_binding_multiple_commands_same_action() -> None:
    """Verify multiple commands bind to an action."""
    scheduler = Scheduler.instance or Scheduler()
    scheduler._reset_all_stacks()

    # Verify the stack is empty
    assert not scheduler._actions_stack

    # Add a new mapping
    class MyAction(Action):

        def poll(self) -> bool:
            return True

    class MyCommand(Command):

        def is_finished(self) -> bool:
            return True

        def execute(self) -> None:
            return None

    action = MyAction()
    command1 = MyCommand()
    command2 = MyCommand()
    condition = Condition.when_activated

    scheduler.bind_command(action, command1, condition)
    assert action in scheduler._actions_stack
    assert scheduler._actions_stack[action] == {condition: {command1}}

    # Add a second mapping
    scheduler.bind_command(action, command2, condition)
    assert action in scheduler._actions_stack
    assert scheduler._actions_stack[action] == {condition: {command1, command2}}

def test_rebinding_multiple_commands_same_action() -> None:
    """Verify rebinding multiple commands on the same action."""
    scheduler = Scheduler.instance or Scheduler()
    scheduler._reset_all_stacks()

    # Verify the stack is empty
    assert not scheduler._actions_stack

    # Add a new mapping
    class MyAction(Action):

        def poll(self) -> bool:
            return True

    class MyCommand(Command):

        def is_finished(self) -> bool:
            return True

        def execute(self) -> None:
            return None

    action = MyAction()
    command1 = MyCommand()
    command2 = MyCommand()
    condition = Condition.when_activated

    scheduler.bind_command(action, command1, condition)
    assert action in scheduler._actions_stack
    assert scheduler._actions_stack[action] == {condition: {command1}}

    # Add a second mapping
    scheduler.bind_command(action, command2, condition)
    assert action in scheduler._actions_stack
    assert scheduler._actions_stack[action] == {condition: {command1, command2}}

    # Rebind the first mapping
    new_condition = Condition.when_deactivated
    scheduler.bind_command(action, command1, new_condition)
    assert action in scheduler._actions_stack
    assert len(scheduler._actions_stack) == 1
    assert len(scheduler._actions_stack[action]) == 2
    assert scheduler._actions_stack[action][condition] == {command2}
    assert scheduler._actions_stack[action][new_condition] == {command1}


def test_binding_multiple_actions() -> None:
    """Verify binding a command to multiple actions."""
    scheduler = Scheduler.instance or Scheduler()
    scheduler._reset_all_stacks()

    # Verify the stack is empty
    assert not scheduler._actions_stack

    # Add a new mapping
    class MyAction(Action):

        def poll(self) -> bool:
            return True

    class MyCommand(Command):

        def is_finished(self) -> bool:
            return True

        def execute(self) -> None:
            return None

    action1 = MyAction()
    action2 = MyAction()
    command = MyCommand()
    condition1 = Condition.when_activated
    condition2 = Condition.when_held

    scheduler.bind_command(action1, command, condition1)
    scheduler.bind_command(action2, command, condition2)
    assert action1 in scheduler._actions_stack
    assert action2 in scheduler._actions_stack
    assert scheduler._actions_stack[action1] == {condition1: {command}}
    assert scheduler._actions_stack[action2] == {condition2: {command}}


def test_cancel_command() -> None:
    """Verify commands are canceled."""
    scheduler = Scheduler.instance or Scheduler()
    scheduler._reset_all_stacks()

    # Verify the stack is empty
    assert not scheduler._actions_stack

    class MyAction(Action):

        def poll(self) -> bool:
            return True

    class MyCommand(Command):

        canceled_counter = 0

        def is_finished(self) -> bool:
            return True

        def execute(self) -> None:
            return None

        def end(self, interrupted: bool) -> None:
            self.canceled_counter += 1
            assert interrupted

    action = MyAction()
    command = MyCommand()

    # Inject the command into the scheduler's stack
    scheduler._all_stack.add(command)

    # Cancel the command
    scheduler.cancel()
    assert command.canceled_counter == 1

    # Cancel again, but verify the command is already de-stacked
    scheduler.cancel()
    assert command.canceled_counter == 1


def test_command_raises_runtime_warning_in_cancel() -> None:
    """Verify commands raise RuntimeWarnings if they fail to cancel."""
    scheduler = Scheduler.instance or Scheduler()
    scheduler._reset_all_stacks()

    # Verify the stack is empty
    assert not scheduler._actions_stack

    class MyCommand(Command):

        def is_finished(self) -> bool:
            return True

        def execute(self) -> None:
            return None

        def end(self, interrupted: bool) -> None:
            raise ValueError("test error")

    command = MyCommand()

    # Artificially inject the command into the stack
    scheduler._all_stack.add(command)

    # Verify the command fails and raises a warning if not explicitly
    # specified
    with pytest.warns(RuntimeWarning):
        scheduler.cancel()

    # Re-inject the command
    scheduler._all_stack.add(command)

    # Verify the command fails and raises a warning if explicitly
    # specified
    with pytest.warns(RuntimeWarning):
        scheduler.cancel(command)


def test_scheduler_event_loop() -> None:
    """Verify the event loop schedules everything correctly."""
    scheduler = Scheduler.instance or Scheduler()
    scheduler._reset_all_stacks()

    # Verify the stack is empty
    assert not scheduler._actions_stack


    class MyAction(Action):

        state = False

        def poll(self) -> bool:
            return self.state


    class MyCommand(Command):

        did_init = 0
        did_exec = 0
        did_end = 0
        did_interrupt = 0
        did_finish = 0
        do_finish = False

        def initialize(self) -> None:
            self.did_init += 1

        def execute(self) -> None:
            self.did_exec += 1

        def end(self, interrupted: bool) -> None:
            if interrupted:
                self.did_interrupt += 1
                return
            self.did_end += 1

        def is_finished(self) -> bool:
            self.did_finish += 1
            return self.do_finish

        def reset(self) -> None:
            self.did_init = 0
            self.did_exec = 0
            self.did_end = 0
            self.did_interrupt = 0
            self.do_finish = False


    class MySubsystem(Subsystem):

        periodic_counter = 0

        def periodic(self) -> None:
            self.periodic_counter += 1

        def reset(self) -> None:
            self.periodic_counter = 0


    action = MyAction()
    subsystem = MySubsystem()
    subsystem2 = MySubsystem()
    command1 = MyCommand("Command1", subsystem)
    command2 = MyCommand("Command2", subsystem)
    command3 = MyCommand("Command3", subsystem)
    command4 = MyCommand("Command4", subsystem)
    command5 = MyCommand("Command5", subsystem)
    command6 = MyCommand("Command6", subsystem2)

    # Bind each command to separate conditions
    action.toggle_when_activated(command2)
    action.when_activated(command3)
    action.when_deactivated(command4)
    action.when_held(command5)
    action.cancel_when_activated(command6)

    # Set the subsystem's default command to command1
    subsystem.default_command = command1

    # Verify the scheduler is tracking the commands correctly
    assert scheduler._actions_stack == {
        action: {
            Condition.cancel_when_activated: {command6},
            Condition.toggle_when_activated: {command2},
            Condition.when_activated: {command3},
            Condition.when_deactivated: {command4},
            Condition.when_held: {command5},
        }
    }

    # Inject command6 as if it was already executing
    scheduler._scheduled_stack.add(command6)

    # Run one loop of the scheduler
    scheduler.run_once()

    # Verify the scheduler has updated
    assert command1.did_init == 1
    assert command1.did_exec == 0
    assert command1.did_end == 0
    assert command1.did_interrupt == 0
    assert command1.did_finish == 0
    assert scheduler._scheduled_stack == {command1, command6}
    assert not scheduler._incoming_stack
    assert subsystem.current_command == command1
    assert subsystem.periodic_counter == 1
    assert command6.did_exec == 1

    # Run again to ensure command1's exec is called
    scheduler.run_once()
    assert command1.did_init == 1
    assert command1.did_exec == 1
    assert command1.did_end == 0
    assert command1.did_interrupt == 0
    assert command1.did_finish == 1
    assert subsystem.current_command == command1
    assert subsystem.periodic_counter == 2
    assert command6.did_exec == 2

    # Now activate the action
    action.state = True

    # Run the scheduler and verify all commands are stacked correctly
    scheduler.run_once()  # when activated

    assert command1.did_init == 1
    assert command1.did_exec == 1
    assert command1.did_end == 0
    assert command1.did_interrupt == 1
    assert command1.did_finish == 1

    assert command2.did_init != command3.did_init
    assert not command2.did_exec and not command3.did_exec
    assert not command2.did_end and not command3.did_end
    assert not command2.did_finish and not command3.did_finish
    assert not command2.did_interrupt and not command3.did_interrupt
    assert subsystem.current_command == (command2 if command2.did_init else command3)
    assert subsystem.periodic_counter == 3

    assert command6.did_exec == 2
    assert command6.did_interrupt == 1

    scheduler.run_once()  # when held

    assert command1.did_init == 1
    assert command1.did_exec == 1
    assert command1.did_end == 0
    assert command1.did_interrupt == 1
    assert command1.did_finish == 1

    assert len({1, 0}.intersection({command2.did_init, command3.did_init})) == 2
    assert not command2.did_exec and not command3.did_exec
    assert not command2.did_end and not command3.did_end
    assert not command2.did_finish and not command3.did_finish
    assert len({1, 0}.intersection({command2.did_interrupt, command3.did_interrupt})) == 2

    assert command5.did_init == 1
    assert command5.did_exec == 0
    assert command5.did_end == 0
    assert command5.did_interrupt == 0
    assert command5.did_finish == 0
    assert subsystem.current_command == command5
    assert subsystem.periodic_counter == 4

    scheduler.run_once()  # when held

    assert command5.did_init == 1
    assert command5.did_exec == 1
    assert command5.did_end == 0
    assert command5.did_interrupt == 0
    assert command5.did_finish == 1
    assert subsystem.periodic_counter == 5

    # Deactivate the action
    action.state = False
    scheduler.run_once()  # when deactivated

    assert command5.did_init == 1
    assert command5.did_exec == 1
    assert command5.did_end == 1
    assert command5.did_interrupt == 0
    assert command5.did_finish == 1

    assert command4.did_init == 1
    assert command4.did_exec == 0
    assert command4.did_end == 0
    assert command4.did_interrupt == 0
    assert command4.did_finish == 0
    assert subsystem.current_command == command4
    assert subsystem.periodic_counter == 6

    scheduler.run_once()

    assert command4.did_init == 1
    assert command4.did_exec == 1
    assert command4.did_end == 0
    assert command4.did_interrupt == 0
    assert command4.did_finish == 1
    assert subsystem.periodic_counter == 7

    # End the command
    command4.do_finish = True
    scheduler.run_once()

    assert command4.did_init == 1
    assert command4.did_exec == 1
    assert command4.did_interrupt == 0
    assert command4.did_end == 1
    assert command4.did_finish == 2
    assert subsystem.current_command == None
    assert subsystem.periodic_counter == 8

    # Create another command and action
    action2 = MyAction()
    command6 = MyCommand("Command6", subsystem)

    action2.when_activated(command6)

    # Verify command1 is rescheduled as the default
    scheduler.run_once()

    assert command1.did_init == 2
    assert command1.did_exec == 1
    assert command1.did_end == 0
    assert command1.did_interrupt == 1
    assert command1.did_finish == 1
    assert subsystem.current_command == command1
    assert subsystem.periodic_counter == 9

    # Interrupt the default command
    action2.state = True
    scheduler.run_once()

    assert command1.did_init == 2
    assert command1.did_exec == 1
    assert command1.did_end == 0
    assert command1.did_interrupt == 2
    assert command1.did_finish == 1
    assert subsystem.current_command == command6
    assert subsystem.periodic_counter == 10


def test_toggle_commands() -> None:
    """Verify commands toggle."""
    scheduler = Scheduler.instance or Scheduler()
    scheduler._reset_all_stacks()

    # Verify the stack is empty
    assert not scheduler._actions_stack


    class MyAction(Action):

        state = False

        def poll(self) -> bool:
            return self.state


    class MyCommand(Command):

        did_init = 0
        did_exec = 0
        did_end = 0
        did_interrupt = 0
        did_finish = 0
        do_finish = False

        def initialize(self) -> None:
            self.did_init += 1

        def execute(self) -> None:
            self.did_exec += 1

        def end(self, interrupted: bool) -> None:
            if interrupted:
                self.did_interrupt += 1
                return
            self.did_end += 1

        def is_finished(self) -> bool:
            self.did_finish += 1
            return self.do_finish

        def reset(self) -> None:
            self.did_init = 0
            self.did_exec = 0
            self.did_end = 0
            self.did_interrupt = 0
            self.do_finish = False

    command = MyCommand()
    action = MyAction()
    action.state = True

    action.toggle_when_activated(command)

    scheduler.run_once()

    assert command.did_init == 1
    assert command.did_exec == 0
    assert command.did_end == 0
    assert command.did_interrupt == 0
    assert command.did_finish == 0

    scheduler.run_once()

    assert command.did_init == 1
    assert command.did_exec == 1
    assert command.did_end == 0
    assert command.did_interrupt == 0
    assert command.did_finish == 1

    action.state = False

    scheduler.run_once()

    assert command.did_init == 1
    assert command.did_exec == 2
    assert command.did_end == 0
    assert command.did_interrupt == 0
    assert command.did_finish == 2

    action.state = True

    scheduler.run_once()

    assert command.did_init == 1
    assert command.did_exec == 2
    assert command.did_end == 1
    assert command.did_interrupt == 0
    assert command.did_finish == 2


def test_command_error_cancels() -> None:
    """Verify commands cancel if methods raise unhandable exceptions."""
    scheduler = Scheduler.instance or Scheduler()
    scheduler._reset_all_stacks()

    # Verify the stack is empty
    assert not scheduler._actions_stack

    class MyCommand(Command):

        interrupted = False

        def is_finished(self) -> bool:
            return False

        def execute(self) -> None:
            raise ValueError("test")

        def end(self, interrupted: bool) -> None:
            self.interrupted = interrupted

        def handle_exception(self, *exc) -> bool:
            return False

    class MyCommand2(Command):

        interrupted = False

        def initialize(self) -> None:
            raise ValueError("test")

        def is_finished(self) -> bool:
            return False

        def execute(self) -> None:
            pass

        def end(self, interrupted: bool) -> None:
            self.interrupted = interrupted

        def handle_exception(self, *exc) -> bool:
            return False

    command = MyCommand()
    command2 = MyCommand2()

    # Inject the command
    scheduler._incoming_stack.add(command2)
    scheduler._all_stack.add(command)
    scheduler._scheduled_stack.add(command)

    scheduler.run_once()

    assert command.interrupted
    assert command2.interrupted


def test_conflicting_incoming_commands() -> None:
    """Verify incoming commands with conflicting requirements."""
    scheduler = Scheduler.instance or Scheduler()
    scheduler._reset_all_stacks()

    # Verify the stack is empty
    assert not scheduler._actions_stack


    class MyCommand(Command):

        did_init = 0
        did_exec = 0
        did_end = 0
        did_interrupt = 0
        did_finish = 0
        do_finish = False

        def initialize(self) -> None:
            self.did_init += 1

        def execute(self) -> None:
            self.did_exec += 1

        def end(self, interrupted: bool) -> None:
            if interrupted:
                self.did_interrupt += 1
                return
            self.did_end += 1

        def is_finished(self) -> bool:
            self.did_finish += 1
            return self.do_finish

        def reset(self) -> None:
            self.did_init = 0
            self.did_exec = 0
            self.did_end = 0
            self.did_interrupt = 0
            self.do_finish = False


    class MySubsystem(Subsystem):

        def periodic(self) -> None:
            pass

    subsystem1 = MySubsystem()
    subsystem2 = MySubsystem()
    command1 = MyCommand("Command1", subsystem1)
    command2 = MyCommand("Command2", subsystem1)
    command3 = MyCommand("Command3", subsystem2)

    # Inject both commands into the incoming stack
    scheduler._incoming_stack.add(command1)
    scheduler._incoming_stack.add(command2)
    scheduler._incoming_stack.add(command3)

    # Verify only one command runs
    scheduler.run_once()
    assert command1.did_init != command2.did_init
    assert command3.did_init


def test_scheduled_incoming_conflicting_commands() -> None:
    """Verify scheduled commands are interrupted by incoming commands."""
    scheduler = Scheduler.instance or Scheduler()
    scheduler._reset_all_stacks()

    # Verify the stack is empty
    assert not scheduler._actions_stack


    class MyCommand(Command):

        did_init = 0
        did_exec = 0
        did_end = 0
        did_interrupt = 0
        did_finish = 0
        do_finish = False

        def initialize(self) -> None:
            self.did_init += 1

        def execute(self) -> None:
            self.did_exec += 1

        def end(self, interrupted: bool) -> None:
            if interrupted:
                self.did_interrupt += 1
                return
            self.did_end += 1

        def is_finished(self) -> bool:
            self.did_finish += 1
            return self.do_finish

        def reset(self) -> None:
            self.did_init = 0
            self.did_exec = 0
            self.did_end = 0
            self.did_interrupt = 0
            self.do_finish = False

    class MySubsystem(Subsystem):

        def periodic(self) -> None:
            pass

    subsystem = MySubsystem()
    subsystem2 = MySubsystem()
    command1 = MyCommand("Command1", subsystem)
    command2 = MyCommand("Command2", subsystem)
    command3 = MyCommand("Command3", subsystem2)

    # Inject one command into the scheduled stack and the other into the
    # incoming stack
    scheduler._scheduled_stack.add(command1)
    scheduler._scheduled_stack.add(command3)
    scheduler._incoming_stack.add(command2)

    # Verify only one command runs
    scheduler.run_once()
    assert not command1.did_exec
    assert command1.did_interrupt
    assert command2.did_init
    assert command3.did_exec


def test_subsystems_dont_default_incoming_commands() -> None:
    """Verify subsystems detect incoming commands and don't default."""
    scheduler = Scheduler.instance or Scheduler()
    scheduler._reset_all_stacks()

    # Verify the stack is empty
    assert not scheduler._actions_stack

    class MyCommand(Command):

        def is_finished(self) -> bool:
            return True

        def execute(self) -> None:
            return None

    class MySubsystem(Subsystem):

        periodic_counter = 0

        def periodic(self) -> None:
            self.periodic_counter += 1

        def reset(self) -> None:
            self.periodic_counter = 0

    subsystem = MySubsystem()
    subsystem2 = MySubsystem()
    command1 = MyCommand("Command1", subsystem)
    command2 = MyCommand("Command2", subsystem)
    command3 = MyCommand("Command3", subsystem2)
    command4 = MyCommand("Command4", subsystem2)

    subsystem.default_command = command1
    subsystem2.default_command = command3

    # Inject the commands to the incoming stack
    scheduler._incoming_stack.add(command2)
    scheduler._incoming_stack.add(command4)

    scheduler.run_once()

    assert subsystem.current_command == command2

    # Cancel command2
    scheduler.cancel(command2)

    assert subsystem.current_command == None


def test_forked_exec() -> None:
    """Verify execute runs normally when forked."""
    scheduler = Scheduler.instance or Scheduler()
    scheduler._reset_all_stacks()

    # Verify the stack is empty
    assert not scheduler._actions_stack

    class MyAction(Action):

        state = False

        def poll(self) -> bool:
            return self.state

    class MyCommand(Command):

        did_exec = False

        def is_finished(self) -> bool:
            return False

        def execute(self) -> None:
            self.did_exec = True
            return None

    class MySubsystem(Subsystem):

        periodic_counter = 0

        def periodic(self) -> None:
            self.periodic_counter += 1

        def reset(self) -> None:
            self.periodic_counter = 0

    action = MyAction()
    subsystem = MySubsystem()
    command1 = MyCommand("Command1", subsystem)

    action.when_activated(command1)

    fut = scheduler.execute(fork=True)

    # Activate the action
    action.state = True

    # Wait 1 second
    time.sleep(1)

    scheduler.shutdown()
    while not fut.done():
        time.sleep(0.1)

    assert command1.did_exec


def test_exec() -> None:
    """Verify execute runs when executed in the main thread."""
    scheduler = Scheduler.instance or Scheduler()
    scheduler._reset_all_stacks()

    # Verify the stack is empty
    assert not scheduler._actions_stack

    class MyAction(Action):

        state = False

        def poll(self) -> bool:
            return self.state

    class MyCommand(Command):

        did_exec = False

        def is_finished(self) -> bool:
            return False

        def execute(self) -> None:
            self.did_exec = True
            return None

    class MySubsystem(Subsystem):

        periodic_counter = 0

        def periodic(self) -> None:
            self.periodic_counter += 1

        def reset(self) -> None:
            self.periodic_counter = 0

    action = MyAction()
    subsystem = MySubsystem()
    command1 = MyCommand("Command1", subsystem)

    action.when_activated(command1)

    def quit_after_one_second(sched: Scheduler) -> None:
        time.sleep(1)
        sched.shutdown()

    thread = Thread(target=quit_after_one_second, args=(scheduler,))
    thread.start()

    # Activate the action
    action.state = True

    # Run the scheduler
    scheduler.execute()

    assert command1.did_exec


@pytest.mark.filterwarnings("ignore::pytest.PytestUnhandledThreadExceptionWarning")
def test_exec_cancels_on_error() -> None:
    """Verify the event loop cancels all commands when errors occur."""
    scheduler = Scheduler.instance or Scheduler()
    scheduler._reset_all_stacks()

    # Verify the stack is empty
    assert not scheduler._actions_stack

    class MyAction(Action):

        state = False

        def poll(self) -> bool:
            return self.state

    class MyCommand(Command):

        did_exec = False
        did_cancel = False

        def is_finished(self) -> bool:
            # Force an error condition
            raise ValueError("test error")

        def execute(self) -> None:
            self.did_exec = True
            return None

        def end(self, interrupted: bool) -> None:
            self.did_cancel = interrupted

    class MySubsystem(Subsystem):

        periodic_counter = 0

        def periodic(self) -> None:
            self.periodic_counter += 1

        def reset(self) -> None:
            self.periodic_counter = 0

    action = MyAction()
    subsystem = MySubsystem()
    command1 = MyCommand("Command1", subsystem)
    command2 = MyCommand("Command2", subsystem)

    action.when_activated(command1)

    fut = scheduler.execute(fork=True)

    # Activate the action
    action.state = True

    # Wait 1 second
    time.sleep(1)

    scheduler.shutdown()
    while not fut.done():
        time.sleep(0.1)

    # command.is_finished is called before command.execute
    assert not command1.did_exec
    assert command1.did_cancel
    assert fut.exception() is not None

    action.when_held(command2)

    def quit_after_one_second(sched: Scheduler) -> None:
        time.sleep(1)
        sched.shutdown()

    thread = Thread(target=quit_after_one_second, args=(scheduler,))
    thread.start()

    # Activate the action
    action.state = True

    # Run the scheduler
    with pytest.raises(ValueError):
        scheduler.execute()

    assert not command2.did_exec
    assert command2.did_cancel


def test_sequential_command_group() -> None:
    """Verify sequential command group operates as expected."""
    scheduler = Scheduler.instance or Scheduler()
    scheduler._reset_all_stacks()

    # Verify the stack is empty
    assert not scheduler._actions_stack

    class MyAction(Action):

        state = False

        def poll(self) -> bool:
            return self.state

    class MyCommand(Command):

        did_init = 0
        did_exec = 0
        did_end = 0
        did_interrupt = 0
        did_finish = 0
        do_finish = False

        def initialize(self) -> None:
            self.did_init += 1

        def execute(self) -> None:
            self.did_exec += 1

        def end(self, interrupted: bool) -> None:
            if interrupted:
                self.did_interrupt += 1
                return
            self.did_end += 1

        def is_finished(self) -> bool:
            self.did_finish += 1
            return self.do_finish

        def reset(self) -> None:
            self.did_init = 0
            self.did_exec = 0
            self.did_end = 0
            self.did_interrupt = 0
            self.do_finish = False

    class MySubsystem(Subsystem):

        periodic_counter = 0

        def periodic(self) -> None:
            self.periodic_counter += 1

        def reset(self) -> None:
            self.periodic_counter = 0

    action = MyAction()
    subsystem = MySubsystem()
    command1 = MyCommand("Command1", subsystem)
    command2 = MyCommand("Command2", subsystem)
    command3 = MyCommand("Command1", subsystem)
    command4 = MyCommand("Command2", subsystem)
    seq_command = SequentialCommandGroup("MySequentialCommandGroup", command1, command2)
    seq_command2 = SequentialCommandGroup("MySequentialCommandGroup2", command3, command4)

    subsystem.default_command = seq_command2

    action.when_activated(seq_command)
    action.state = True

    scheduler.run_once()

    assert subsystem.current_command == seq_command
    assert command1.did_init == 1
    assert command2.did_init == 0

    scheduler.run_once()

    assert command1.did_init == 1
    assert command1.did_finish == 1
    assert command1.did_exec == 1
    assert command2.did_init == 0

    command1.do_finish = True

    scheduler.run_once()

    assert command1.did_exec == 1
    assert command1.did_finish == 2
    assert command1.did_end == 1
    assert command2.did_init == 1

    scheduler.run_once()

    assert command2.did_init == 1
    assert command2.did_exec == 1
    assert command2.did_finish == 1

    command2.do_finish = True

    scheduler.run_once()
    scheduler.run_once()

    assert command2.did_exec == 1
    assert command2.did_finish == 2
    assert command2.did_end == 1
    assert seq_command._end_of_sequence
    assert subsystem.current_command == None

    scheduler.run_once()

    assert subsystem.current_command == seq_command2
    assert command3.did_init == 1

    scheduler.run_once()

    assert command3.did_exec == 1
    assert command3.did_finish == 1

    # Cancel the sequential command
    subsystem.default_command = None
    scheduler.cancel(seq_command2)

    assert subsystem.current_command == None
    assert command3.did_interrupt == 1
    assert command4.did_init == 0
    assert command4.did_exec == 0
    assert command4.did_end == 0
    assert command4.did_finish == 0
    assert command4.did_interrupt == 0


def test_parallel_command_group() -> None:
    """Verify parallel command groups execute as expected."""
    scheduler = Scheduler.instance or Scheduler()
    scheduler._reset_all_stacks()

    # Verify the stack is empty
    assert not scheduler._actions_stack

    class MyAction(Action):

        state = False

        def poll(self) -> bool:
            return self.state

    class MyCommand(Command):

        did_init = 0
        did_exec = 0
        did_end = 0
        did_interrupt = 0
        did_finish = 0
        do_finish = False
        raise_error = False

        def handle_exception(self, *exc) -> bool:
            print(exc)
            return True

        def initialize(self) -> None:
            self.did_init += 1

        def execute(self) -> None:
            self.did_exec += 1
            if self.raise_error:
                raise ValueError("test")

        def end(self, interrupted: bool) -> None:
            if interrupted:
                self.did_interrupt += 1
                return
            self.did_end += 1

        def is_finished(self) -> bool:
            self.did_finish += 1
            return self.do_finish

        def reset(self) -> None:
            self.did_init = 0
            self.did_exec = 0
            self.did_end = 0
            self.did_interrupt = 0
            self.do_finish = False

    class MySubsystem(Subsystem):

        periodic_counter = 0

        def periodic(self) -> None:
            self.periodic_counter += 1

        def reset(self) -> None:
            self.periodic_counter = 0

    action = MyAction()
    action2 = MyAction()
    subsystem = MySubsystem()
    subsystem2 = MySubsystem()
    command1 = MyCommand("Command1", subsystem)
    command2 = MyCommand("Command2", subsystem2)
    command3 = MyCommand("Command3", subsystem)
    command4 = MyCommand("Command4", subsystem2)
    command5 = MyCommand("Command5", subsystem)
    command6 = MyCommand("Command6", subsystem2)
    par_command = ParallelCommandGroup("MyParallelCommandGroup", command1, command2)
    par_command2 = ParallelCommandGroup("MyParallelCommandGroup2", command3, command4)
    par_command3 = ParallelCommandGroup("MyParallelCommandGroup3", command5, command6)

    with pytest.raises(ValueError):
        ParallelCommandGroup("MyParallelCommandGroup", command1, command3)

    action.when_activated(par_command)
    action.state = True

    scheduler.run_once()

    # Allow the threads to execute for a little bit
    time.sleep(1)

    command1.do_finish = True
    command2.do_finish = True

    # Allow parallel threads to reach their barrier
    time.sleep(0.1)

    # Finish executing the parallel command and clean it up
    scheduler.run_once()

    assert command1.did_init == 1
    assert command1.did_exec > 0
    assert command1.did_finish > 0
    assert command1.did_end == 1

    assert command2.did_init == 1
    assert command2.did_exec > 0
    assert command2.did_finish > 0
    assert command2.did_end == 1

    action.when_held(par_command2)

    scheduler.run_once()

    # Allow the threads to execute for a little bit
    time.sleep(1)

    action.state = False

    scheduler.run_once()

    # Allow parallel threads to reach their barrier
    time.sleep(1)

    # Finish executing the parallel command and clean it up
    scheduler.run_once()

    assert command3.did_init == 1
    assert command3.did_exec > 0
    assert command3.did_finish > 0
    assert command3.did_end == 0
    assert command3.did_interrupt == 1

    assert command4.did_init == 1
    assert command4.did_exec > 0
    assert command4.did_finish > 0
    assert command4.did_end == 0
    assert command4.did_interrupt == 1

    action2.when_activated(par_command3)
    action2.state = True

    scheduler.run_once()

    # All threads to execute for a little bit
    time.sleep(1)

    # Raise an error in command 5
    command5.raise_error = True

    time.sleep(20)

    # Ensure the group catches the error and handles it properly
    scheduler.run_once()

    assert command5.did_init == 1
    assert command5.did_exec > 0
    assert command5.did_finish > 0
    assert command5.did_end == 0
    assert command5.did_interrupt == 1

    assert command6.did_init == 1
    assert command6.did_exec > 0
    assert command6.did_finish > 0
    assert command6.did_end == 0
    assert command6.did_interrupt == 1



from abc import ABC, abstractmethod

from command_based_framework.scheduler import Scheduler


class CommandBasedRobot(ABC, Scheduler):
    """A template to ease the creation of a compatible system.

    This template breaks out all major aspects of the framework into
    clearly defined methods of ease-of-use. There is a specific order
    in which these methods are executed.
    """

    @abstractmethod
    def bind_components(self) -> None:
        """Bind all action `whens` to commands in this method."""

    @abstractmethod
    def create_actions(self) -> None:
        """Define and instantiate actions in this method."""

    @abstractmethod
    def create_commands(self) -> None:
        """Define and instantiate commands in this method."""

    @abstractmethod
    def create_inputs(self) -> None:
        """Define and instantiate any inputs in this method."""

    @abstractmethod
    def create_subsystems(self) -> None:
        """Define and instantiate subsystems in this method."""

    def prestart_setup(self) -> None:
        """Called just before the event loop starts.

        Handles the calling of all `create` methods within this class.
        The order in which these are called is as follows:

        **1.** :py:meth:`~command_based_framework.robot.CommandBasedRobot.create_inputs`

        **2.** :py:meth:`~command_based_framework.robot.CommandBasedRobot.create_subsystems`

        **3.** :py:meth:`~command_based_framework.robot.CommandBasedRobot.create_commands`

        **4.** :py:meth:`~command_based_framework.robot.CommandBasedRobot.create_actions`

        **5.** :py:meth:`~command_based_framework.robot.CommandBasedRobot.bind_components`
        """  # noqa: E501
        self.create_inputs()
        self.create_subsystems()
        self.create_commands()
        self.create_actions()
        self.bind_components()

    def postend_teardown(self) -> None:
        """Called just after the event loop ends.

        Any post-execution code should go here.
        """



import itertools
import sys
import time
from abc import ABC, abstractmethod
from concurrent.futures import ThreadPoolExecutor
from threading import Barrier, Event
from types import TracebackType
from typing import Iterator, Optional, Set, Tuple, Type, Union

if sys.version_info >= (3, 10):
    # WPS433: Found nested import
    # WPS440: Found block variables overlap
    from typing import TypeAlias  # noqa: WPS433, WPS440; pragma: no cover
else:
    # WPS433: Found nested import
    # WPS440: Found block variables overlap
    from typing_extensions import TypeAlias  # noqa: WPS433, WPS440; pragma: no cover

from command_based_framework._common import ContextManagerMixin
from command_based_framework.scheduler import Scheduler
from command_based_framework.subsystems import Subsystem

CommandType: TypeAlias = Union["Command", "CommandGroup"]


class Command(ABC, ContextManagerMixin):  # noqa: WPS214
    """Executes a process when activated by an :py:class:`~command_based_framework.actions.Action`.

    Commands dictate what subsystems do at what time. They are scheduled
    when a :py:meth:`~command_based_framework.actions.Action.poll`
    bound condition is met. Commands are also synchronous, meaning they
    are always blocking the scheduler's event loop and should complete
    quickly.

    Commands have the following life cycle in the scheduler:

    **1.** New commands have their :py:meth:`~command_based_framework.commands.Command.initialize`
    method called.

    **2.** Actions bound to this command have their :py:meth:`~command_based_framework.actions.Action.poll`
    method called. Depending on how a command is bound to an action, the
    scheduler may skip directly to step 4 for a command.

    **3.** The scheduler now periodically executes these new commands by
    calling their :py:meth:`~command_based_framework.commands.Command.is_finished`
    and :py:meth:`~command_based_framework.commands.Command.execute`
    methods, in that order.

    **4.** Whether through an action or :py:meth:`~command_based_framework.commands.Command.is_finished`,
    commands have their :py:meth:`~command_based_framework.commands.Command.end`
    methods called and are removed from the stack.

    Commands also maintain their state after being unscheduled as long
    as a reference is maintained. The scheduler maintains a reference as
    long as the command is scheduled, but releases it immediately after.
    """  # noqa: E501

    # The name of the command
    _name: str

    # Requirements are the subsystems required for this command to run.
    # The scheduler uses this to ensure only one command is using a
    # subsystem at any time
    _requirements: Set[Subsystem]

    # Indicates whether or not the command needs to be interrupted after
    # encountering an error
    _needs_interrupt: bool

    def __init__(self, name: Optional[str] = None, *subsystems: Subsystem) -> None:
        """Creates a new `Command` instance.

        :param name: The name of the command. If not provided, the class
            name is used instead.
        :type name: str, optional
        :param subsystems: Variable length of subsystems to
            automatically require.
        :type subsystems: tuple
        """
        super().__init__()
        self._name = name or self.__class__.__name__
        self._requirements = set()
        self._needs_interrupt = False

        # Register each subsystem as a requirements
        self.add_requirements(*subsystems)

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]] = None,
        exc: Optional[BaseException] = None,
        traceback: Optional[TracebackType] = None,
    ) -> bool:
        """Called when the command exits a context manager."""
        # Ignore non-errors
        if not exc_type or not exc or not traceback:
            return True

        handled = self.handle_exception(exc_type, exc, traceback)
        self.needs_interrupt = not handled or not isinstance(handled, bool)
        return True

    def __repr__(self) -> str:
        """Unicode representation of the class."""
        return self.__str__()

    def __str__(self) -> str:
        """Unicode representation of the class."""
        return "<{name}>".format(name=self.name)

    @property
    def name(self) -> str:
        """The name of the command.

        This is a read-only property.

        If one was not provided at the creation of the command, the
        class name is used instead.
        """
        return self._name

    @property
    def needs_interrupt(self) -> bool:
        """Indicates if the command needs to be interrupted.

        This property should not be set directly as it is managed by the
        scheduler.

        Every read of this property resets its state.
        """
        ret = self._needs_interrupt
        self.needs_interrupt = False
        return ret

    @needs_interrupt.setter
    def needs_interrupt(self, state: bool) -> None:
        self._needs_interrupt = state

    @property
    def requirements(self) -> Set[Subsystem]:
        """The subsystems this command requires to run.

        This is a read-only property.
        """
        return self._requirements

    def add_requirements(self, *subsystems: Subsystem) -> None:
        """Register any number of subsystems as a command requirement.

        Only one command can be running with any given requirement. If
        two commands share any requirement and are scheduled to run,
        which command runs may be undefined. If one command is already
        scheduled then it will be interrupted by the newly scheduled
        command.
        """
        self._requirements.update(set(subsystems))

    def handle_exception(
        self,
        exc_type: Type[BaseException],
        exc: BaseException,
        traceback: TracebackType,
    ) -> bool:
        """Called when :py:meth:`~command_based_framework.commands.Command.execute` raises an error.

        The scheduler uses the output of this method to determine
        whether the command should be immediately interrupted.

        :param exc_type: The type of exception raised.
        :type exc_type: :py:class:`Type`
        :param exc: The exception raised.
        :type exc: :py:class:`Exception`
        :param traceback: The frame traceback of the error.
        :type traceback: :py:class:`Traceback`

        :return: `True` to indicate the error is handled. All other
            returns to the scheduler will be interpreted as the command
            needing to be immediately interrupted.
        :rtype: bool
        """  # noqa: DAR202

    def initialize(self) -> None:
        """Called each time the command in scheduled.

        Any initialization or pre-execution code should go here.
        """

    @abstractmethod
    def execute(self) -> None:
        """Periodically called while the command is scheduled.

        All execution code should go here.
        """

    def end(self, interrupted: bool) -> None:
        """Called once the command has been unscheduled.

        Any clean up or post-execution code should go here.

        :param interrupted: the command was interrupted, not ended.
        :type interrupted: bool
        """

    @abstractmethod
    def is_finished(self) -> bool:
        """Periodically called before :py:meth:`~command_based_framework.commands.Command.execute` while the command is scheduled.

        :return: `True` if the command should end, otherwise `False`.
        :rtype: bool
        """  # noqa: E501


class CommandGroup(Command):
    """Group commands into a single manageable interface.

    Only provides a modified command `__init__` method which accepts
    commands instead of subsystems. Also requires all subsystems by any
    passed commands which may share requirements. Note this behavior may
    change on child classes of this abstract.

    Any command or group can be provided to this abstract.
    """

    _commands: Tuple[CommandType, ...]

    def __init__(self, name: Optional[str] = None, *commands: CommandType) -> None:
        """Creates a new :class:`~CommandGroup` instance.

        Args:
            name: The name of the command group. If not provided, the
                class name is used instead.
            commands: Variable length of commands to group.
        """  # noqa: RST203
        super().__init__(name)

        # Require all subsystems
        for command in commands:
            self.add_requirements(*command.requirements)

        self._commands = commands


class SequentialCommandGroup(CommandGroup):
    """Sequentially executes commands.

    The order which commands are provided to the constructor determines
    the order the commands are executed. If any commands error or are
    interrupted, no further commands will be executed.

    Any command or group can be provided to the constructor.
    """

    _sequence: Iterator[CommandType]
    _current_command: Optional[CommandType]
    _end_of_sequence: bool

    def __init__(self, name: Optional[str] = None, *commands: CommandType) -> None:
        """Creates a new :class:`~SequentialCommandGroup` instance.

        Args:
            name: The name of the command group. If not provided, the
                class name is used instead.
            commands: Variable length of commands to group.
        """  # noqa: RST203
        super().__init__(name, *commands)
        self._end_of_sequence = False

    def initialize(self) -> None:
        """Select the first command in the chain to run."""
        self._end_of_sequence = False
        self._sequence = iter(self._commands)
        self._prepare_next_command()

    def execute(self) -> None:
        """Execute the currently sequenced command."""
        # mypy fix, check if self._current_command is set
        if not self._current_command:  # pragma: no cover
            self._end_of_sequence = True  # pragma: no cover
            return  # pragma: no cover

        # Check if the current command has finished
        # If so, end it and prepare the next one
        if self._current_command.is_finished():
            self._current_command.end(interrupted=False)
            self._prepare_next_command()
            return

        # Execute the current command
        self._current_command.execute()

    def is_finished(self) -> bool:
        """Check if the end of the chain has been reached."""
        return self._end_of_sequence or not self._current_command

    def end(self, interrupted: bool) -> None:
        """End the current command."""
        if self._current_command:
            self._current_command.end(interrupted=interrupted)

    def _prepare_next_command(self) -> None:
        try:
            self._current_command = next(self._sequence)
        except StopIteration:
            self._current_command = None
            self._end_of_sequence = True
            return

        # Initialize the command
        self._current_command.initialize()


class ParallelCommandGroup(CommandGroup):
    """Run multiple commands in parallel.

    Each command will execute in its own dedicated thread. Unlike most
    other command groups, commands submitted here cannot share
    requirements.
    """

    _pool: ThreadPoolExecutor
    _finished: Barrier
    _sentinel: Event

    def __init__(  # noqa: WPS231
        self,
        name: Optional[str] = None,
        *commands: CommandType,
    ) -> None:
        """Creates a new :py:class:`~ParallelCommandGroup` instance.

        Args:
            name: The name of the command group. If not provided, the
                class name is used instead.
            commands: Variable length of commands to group.
        """  # noqa: RST203
        # Don't pass the commands to the parent class, we have to
        # implement custom checking of them here
        super().__init__(name)

        # Check for shared requirements
        for command in commands:
            for cmd in commands:
                if cmd == command:
                    continue
                if cmd.requirements.intersection(command.requirements):
                    raise ValueError(
                        "{cmd1} has shared requirements with {cmd2}".format(
                            cmd1=command,
                            cmd2=cmd,
                        ),
                    )
            self.add_requirements(*command.requirements)

        self._commands = commands
        self._sentinel = Event()

    def initialize(self) -> None:
        """Submit all commands to the thread pool."""
        self._sentinel.clear()
        self._pool = ThreadPoolExecutor(len(self._commands) + 1)
        self._finished = Barrier(len(self._commands) + 1)
        self._pool.map(
            self._execute,
            self._commands,
            itertools.repeat(self._finished),
            itertools.repeat(self._sentinel),
        )

    def execute(self) -> None:
        """Not implemented."""

    def is_finished(self) -> bool:
        """Check if all commands have finished."""
        # Do not catch BrokenPipeErrors
        # We want to ensure commands are interrupted in the event
        # something unexpected occurs
        return self._finished.n_waiting == len(self._commands)

    def end(self, interrupted: bool) -> None:
        """Wait for all commands to finish executing."""
        self._sentinel.set()
        self._finished.wait()
        self._pool.shutdown(wait=True)

    def _execute(
        self,
        command: CommandType,
        finished: Barrier,
        sentinel: Event,
    ) -> None:
        try:  # noqa: WPS229
            # Initialize the command
            command.initialize()

            # Run an event loop
            while True:
                # Check if the current command has finished
                if command.is_finished():
                    command.end(interrupted=False)
                    return

                # Check if the thread pool is shutting down
                if sentinel.is_set():
                    command.end(interrupted=True)
                    return

                # Execute the current command
                command.execute()

                time.sleep(Scheduler.instance.clock_speed)  # type: ignore
        except Exception:
            # Handle errors
            command.handle_exception(*sys.exc_info())
            command.end(interrupted=True)

            # Stop execution
            sentinel.set()
        finally:
            finished.wait()



from command_based_framework.actions import Action, Condition
from command_based_framework.commands import Command
from command_based_framework.scheduler import Scheduler

def test_bind_all_condition_types() -> None:
    """Verify all condition types bind properly."""
    scheduler = Scheduler.instance or Scheduler()
    scheduler._reset_all_stacks()

    # Verify the stack is empty
    assert not scheduler._actions_stack

    class MyAction(Action):

        def poll(self) -> bool:
            return True

    class MyCommand(Command):

        def is_finished(self) -> bool:
            return True

        def execute(self) -> None:
            return None

    action = MyAction()
    command_cancel_when_activated = MyCommand()
    command_toggle_when_activated = MyCommand()
    command_when_activated = MyCommand()
    command_when_held = MyCommand()
    command_when_deactivated = MyCommand()

    action.cancel_when_activated(command_cancel_when_activated)
    action.toggle_when_activated(command_toggle_when_activated)
    action.when_activated(command_when_activated)
    action.when_deactivated(command_when_deactivated)
    action.when_held(command_when_held)

    assert scheduler._actions_stack[action][Condition.cancel_when_activated] == {command_cancel_when_activated}
    assert scheduler._actions_stack[action][Condition.toggle_when_activated] == {command_toggle_when_activated}
    assert scheduler._actions_stack[action][Condition.when_activated] == {command_when_activated}
    assert scheduler._actions_stack[action][Condition.when_deactivated] == {command_when_deactivated}
    assert scheduler._actions_stack[action][Condition.when_held] == {command_when_held}



import sys

if sys.version_info[:2] >= (3, 8):
    from importlib.metadata import PackageNotFoundError, version
else:
    from importlib_metadata import PackageNotFoundError, version

dist_name = "command-based-framework"
try:
    __version__ = version(dist_name)
except PackageNotFoundError:
    __version__ = "unknown"
finally:
    del version, PackageNotFoundError



import sys
import time
import warnings
import weakref
from abc import ABCMeta
from concurrent.futures import Future
from contextlib import suppress
from threading import Event, Thread
from typing import Dict, Optional, Set

if sys.version_info >= (3, 10):
    # WPS433: Found nested import
    # WPS440: Found block variables overlap
    from typing import TypeAlias  # pragma: no cover
else:
    # WPS433: Found nested import
    # WPS440: Found block variables overlap
    from typing_extensions import TypeAlias  # pragma: no cover

# Annotations only
with suppress(ImportError):
    from command_based_framework.actions import Action, Condition  # pragma: no cover
    from command_based_framework.commands import Command  # pragma: no cover
    from command_based_framework.subsystems import Subsystem  # pragma: no cover

from command_based_framework.exceptions import SchedulerExistsError

ConditionCommandType: TypeAlias = Dict["Condition", Set["Command"]]
ActionStack: TypeAlias = Dict["Action", ConditionCommandType]


class SchedulerMeta(ABCMeta, type):  # noqa: B024
    """Meta attributes for :py:class:`~command_based_framework.scheduler.Scheduler`.

    All methods and attributes here exist in :py:class:`~command_based_framework.scheduler.Scheduler`.
    """  # noqa: E501

    _instance: Optional["weakref.ReferenceType[Scheduler]"]

    @property
    def instance(cls) -> Optional["Scheduler"]:
        """The class-level instance attribute.

        Ensures that no more than one scheduler instance exists else
        undefined behavior regarding subsystems may occur.

        :return: The :py:class:`~command_based_framework.scheduler.Scheduler`
            instance if set, otherwise `None`.

        :raise SchedulerExistsError:
            An attempt was made to create multiple :py:class:`~command_based_framework.scheduler.Scheduler`
            instances.
        """  # noqa: E501
        if cls._instance is None:
            return None

        return cls._instance()

    @instance.setter
    def instance(cls, instance: "Scheduler") -> None:
        # Verify no previous instance exists
        if Scheduler.instance:
            raise SchedulerExistsError(
                "a scheduler already exists, a new one cannot be created",
            )
        cls._instance = weakref.ref(instance)


class Scheduler(object, metaclass=SchedulerMeta):
    """Schedules :py:class:`~command_based_framework.commands.Command` that control :py:class:`~command_based_framework.subsystems.Subsystem` when activated by :py:class:`~command_based_framework.actions.Action`.

    The scheduler handles events and resource management for the
    framework. A scheduler **must** exist **before** creating any
    commands, subsystems, or actions. Only one scheduler can exist in a
    program at any given time. More than one scheduler will result in
    undefined behavior.

    The scheduler has the following life-cycle:

    **1**: All actions with bound `when` methods have their :py:meth:`~command_based_framework.actions.Action.poll`
    methods called. For efficiency, unbound actions are ignored.

    **2**: If a `when` condition is met, the related command(s) are
    put onto the incoming stack. Any two commands which share
    requirements (subsystems) will result in the currently scheduled
    command being interrupted. Two newly scheduled commands with shared
    requirements will result in only one of the commands being scheduled
    and a warning being thrown about the conflict.

    **3**: Any subsystem not bound to a scheduled or incoming command
    will have their default command scheduled, if there is one.

    **4**: Commands scheduled for interruption are interrupted. Commands
    scheduled to exit because their :py:meth:`~command_based_framework.commands.Command.is_finished`
    returned `True`, are ended. Incoming commands are initialized. Note
    that commands initialized in one frame will not be normally executed
    until the next frame.

    **5**: Currently scheduled commands have their :py:meth:`~command_based_framework.commands.Command.is_finished`
    methods checked. If `True` is returned, the command(s) are scheduled
    for exit and their :py:meth:`~command_based_framework.commands.Command.execute`
    method is ignored. Otherwise, their :py:meth:`~command_based_framework.commands.Command.execute`
    method is called. Commands are allowed to raise errors from their
    :py:meth:`~command_based_framework.commands.Command.execute` methods
    and their :py:meth:`~command_based_framework.commands.Command.handle_exception`
    method will be called with the output from :py:meth:`sys.exec_info`.
    Return `True` to indicate the error is handled and normal execution
    can continue. Any other return will result in the command being
    immediately interrupted and de-stacked. A warning about the error
    will be thrown.

    **6**: The main stack updates with interrupted/finished commands
    taken off and initialized commands put on.

    **7**: This cycle repeats.

    **8**: Upon shutdown, all current commands are interrupted and
    de-stacked. The scheduler then exits its event loop.
    """  # noqa: E501

    _instance: Optional["weakref.ReferenceType[Scheduler]"] = None

    # Clock speed is how fast the scheduler runs per second
    _clock_speed: float

    # All stack has references to all commands in any stack, regardless
    # of status
    _all_stack: Set["Command"]

    # Action stack has references to the mappings between commands and
    # actions
    _actions_stack: ActionStack

    # Incoming stack has references to all commands that were just
    # scheduled
    _incoming_stack: Set["Command"]

    # Scheduled stack has references to all commands that are normally
    # executing
    _scheduled_stack: Set["Command"]

    # Ended stack has references to all commands that need to be ended
    # normally
    _ended_stack: Set["Command"]

    # Cancel stack has references to all commands that need to be
    # canceled
    _cancel_stack: Set["Command"]

    # Subsystem stack has references to all subsystems that need to
    # have their periodic methods called
    _subsystem_stack: Set["Subsystem"]

    # The thread managing the execution of the event loop
    _exec_thread: Thread

    # Responsible for killing the event loop if it is forked
    _exec_sentinel: Event

    def __init__(self) -> None:
        """Creates a new :py:class:`~command_based_framework.scheduler.Scheduler` instance."""
        Scheduler.instance = self
        self.clock_speed = 1 / 60
        self._reset_all_stacks()
        self._exec_sentinel = Event()

    @property
    def clock_speed(self) -> float:
        """How many times the scheduler will attempt to run per second.

        Because the scheduler is synchronous, long-running commands may
        degrade the ability for the scheduler to stick to this rate.
        This value must always remain above 0 otherwise the CPU may be
        deadlocked.

        Defaults to 60 ticks per second.

        :raise ValueError: An attempt to set the clock speed at or below
            0 was made.
        """
        return self._clock_speed

    @clock_speed.setter
    def clock_speed(self, clock_speed: float) -> None:
        # Ensure the new speed is above 0
        if clock_speed <= 0:
            raise ValueError("clock speed must be at or above 0")

        self._clock_speed = clock_speed

    def bind_command(
        self,
        action: "Action",
        command: "Command",
        condition: "Condition",
    ) -> None:
        """Bind `command` to `action` to be scheduled on `condition`."""
        current_condition_stack = self._actions_stack.setdefault(
            action,
            {condition: {command}},
        )
        for cond, cmdlist in current_condition_stack.items():
            for cmd in cmdlist.copy():
                if cmd == command:
                    cmdlist.remove(cmd)
                    break
            current_condition_stack[cond] = cmdlist
        current_condition_stack.setdefault(condition, set()).add(command)
        self._actions_stack[action] = current_condition_stack
        self._all_stack.add(command)

    def cancel(self, *commands: "Command") -> None:  # noqa: WPS213
        """Immediately cancel and interrupt any number of commands.

        If `commands` is not provided, interrupt all scheduled and
        incoming commands. The `interrupt` parameter of the
        :py:meth:`~command_based_framework.commands.Command.end` method
        for each command will be `True`.

        :param commands: Variable length of commands to cancel. If not
            provided, interrupt all active commands.
        :type commands: tuple
        """
        cancel_all = not commands
        all_commands = set(commands) or self._all_stack
        for command in all_commands.copy():
            try:
                command.end(interrupted=True)
            except Exception:
                command.handle_exception(*sys.exc_info())
                warnings.warn(
                    (
                        "{name} failed to interrupt, this command may have "
                        "failed to properly quit."
                    ).format(name=command.name),
                    RuntimeWarning,
                )

                # Reset the interrupt flag
                command.needs_interrupt  # noqa: WPS428

            # Reset all requirements' current commands to none
            for requirement in command.requirements:
                requirement.current_command = None

            # Remove the command from all stacks
            with suppress(KeyError):
                self._incoming_stack.remove(command)
            with suppress(KeyError):
                self._scheduled_stack.remove(command)
            with suppress(KeyError):
                self._ended_stack.remove(command)

        if cancel_all:
            self._reset_all_stacks()

    def execute(self, fork: bool = False) -> Optional[Future]:
        """Perpetually run the event loop.

        :param fork: Fork a separate thread to run the event loop in. If
            `True`, a :py:class:`~concurrent.futures.Future` is
            returned. If `False` and an attempt is made to shut the
            event loop down via :py:meth:`~command_based_framework.scheduler.Scheduler.shutdown`
            in the same thread, a deadlock will occur.
        :type fork: bool

        :return: A :py:class:`~concurrent.futures.Future` if `fork` is
            `True`, otherwise `None`.
        :rtype: :py:class:`~concurrent.futures.Future`, None
        """  # noqa: DAR202
        # Reset the sentinel
        self._exec_sentinel.clear()

        # Create a fork if necessary
        if fork:
            future: Future = Future()
            self._exec_thread = Thread(target=self._execute, args=(future,))
            self._exec_thread.start()
            return future

        return self._execute()  # type: ignore

    def prestart_setup(self) -> None:
        """Run prestart checks and setup when :py:meth:`~command_based_framework.scheduler.Scheduler.execute` is called."""  # noqa: E501

    def postend_teardown(self) -> None:
        """Run post-end code and teardown when :py:meth:`~command_based_framework.scheduler.Scheduler.execute` exits."""  # noqa: E501

    def register_subsystem(self, subsystem: "Subsystem") -> None:
        """Register a :py:class:`~command_based_framework.subsystems.Subsystem` with the scheduler.

        This should be called automatically by the subsystem upon
        creation, so calling this directly should not be necessary.
        Registered subsystems allow for default commands to be scheduled
        if the subsystem is not active in another command.

        :param subsystem: The subsystem to register.
        :type subsystem: :py:class:`~command_based_framework.subsystems.Subsystem`
        """  # noqa: E501
        self._subsystem_stack.add(subsystem)

    def run_once(self) -> None:
        """Run one complete loop of the scheduler's event loop.

        Note this does not call :py:meth:`~Scheduler.prestart_setup` or
        :py:meth:`~Scheduler.postend_teardown`.
        """
        self._poll_actions()
        self._schedule_default_commands()
        self._end_commands()
        self._init_commands()
        self._execute_commands()
        self._execute_subsystems()
        self._update_stack()

    def shutdown(self) -> None:
        """Shut the scheduler down.

        Any active commands will be interrupted when this method is
        called.
        """
        # Signal the event loop to quit
        self._exec_sentinel.set()

        # AttributeError will be raised if exec has not been forked
        with suppress(AttributeError):
            if self._exec_thread.is_alive():
                self._exec_thread.join()

    def _execute(self, fut: Optional[Future] = None) -> None:  # noqa: C901, WPS231
        # Indicate the thread is running
        if fut:
            fut.set_running_or_notify_cancel()

        try:  # noqa: WPS229
            # Run user-defined code before this function starts
            self.prestart_setup()

            # Main loop
            while not self._exec_sentinel.is_set():
                self.run_once()
                time.sleep(self.clock_speed)
        except Exception as exc:
            # Ensure the parent thread receives the exception
            if fut:
                fut.set_exception(exc)
            raise
        finally:
            # Cancel all commands
            self.cancel()

            # Run postend user-defined code
            self.postend_teardown()

            # Set the future as finished
            if fut:
                fut.set_result(None)

    def _end_commands(self) -> None:
        # End commands normally
        for cmd in self._ended_stack:
            with cmd:
                cmd.end(interrupted=False)
            self._scheduled_stack.remove(cmd)

            # Reset all requirements' current commands to none
            for requirement in cmd.requirements:
                requirement.current_command = None

            # Reset the interrupt flag
            cmd.needs_interrupt  # noqa: WPS428

        # Cancel commands
        if self._cancel_stack:
            self.cancel(*self._cancel_stack)

    def _execute_commands(self) -> None:
        for command in self._scheduled_stack.copy():
            # Check if the command is finished before executing it
            # Safer to do this check first
            if command.is_finished():
                with command:
                    command.end(interrupted=False)
                self._scheduled_stack.remove(command)

                # Reset all requirement's current commands to none
                for requirement in command.requirements:
                    requirement.current_command = None

                # Reset the interrupt flag
                command.needs_interrupt  # noqa: WPS428
                self._ended_stack.add(command)
                continue

            # Execute the command
            with command:
                command.execute()

            # Immediately interrupt the command if needed
            if command.needs_interrupt:
                self.cancel(command)

    def _execute_subsystems(self) -> None:
        for subsystem in self._subsystem_stack:
            with subsystem:
                subsystem.periodic()

    def _init_commands(self) -> None:  # noqa: C901, WPS231
        for command in self._incoming_stack.copy():
            # If the command is already scheduled, don't init
            if command in self._scheduled_stack:
                self._incoming_stack.remove(command)
                continue

            # If the command shares requirements with other commands in
            # the stack, don't init
            skip = False
            for cmd in self._incoming_stack.copy():
                if cmd == command:
                    continue

                if cmd.requirements.intersection(command.requirements):
                    self._incoming_stack.remove(command)
                    skip = True
            if skip:
                continue

            # Interrupt other commands that use this command's
            # requirements
            for cmd in self._scheduled_stack.copy():
                if cmd.requirements.intersection(command.requirements):
                    # Cancel should automatically remove this cmd from
                    # all stacks
                    self.cancel(cmd)

            # Set all requirements current command to this incoming
            # command
            for requirement in command.requirements:
                requirement.current_command = command

            # Initialize this command
            with command:
                command.initialize()

            if command.needs_interrupt:
                # Cancel should automatically remove this cmd from
                # all stacks
                self.cancel(command)

    def _poll_actions(self) -> None:  # noqa: C901, WPS210, WPS231
        # Reimport Condition again since the top-level import will have
        # errored out and is only used for type-hints
        from command_based_framework.actions import Condition  # noqa: WPS442

        for action, conditions_commands in self._actions_stack.items():
            # Get the last and current state of the action
            # Update the last_state immediately since we are checking
            # the current state and the state may change
            last_state = action.last_state
            current_state = action.poll()
            action.last_state = current_state
            action_state = (last_state, current_state)

            # Handle each state type
            if action_state == (False, True):
                # Cancel when activated
                commands = conditions_commands.setdefault(
                    Condition.cancel_when_activated,
                    set(),
                )
                self._cancel_stack.update(commands)

                # When activated
                commands = conditions_commands.setdefault(
                    Condition.when_activated,
                    set(),
                )
                self._incoming_stack.update(commands)

                # Toggle when activated
                commands = conditions_commands.setdefault(
                    Condition.toggle_when_activated,
                    set(),
                )
                for command in commands:
                    if command in self._scheduled_stack:
                        self._ended_stack.add(command)
                    else:
                        self._incoming_stack.add(command)
            elif action_state == (True, True):
                # When held
                commands = conditions_commands.setdefault(Condition.when_held, set())
                self._incoming_stack.update(commands)
            elif action_state == (True, False):
                # When deactivated
                commands = conditions_commands.setdefault(
                    Condition.when_deactivated,
                    set(),
                )
                self._incoming_stack.update(commands)

                # When held
                # Need to be able to deactivate when held, which occurs
                # at the same time as when deactivated
                commands = conditions_commands.setdefault(Condition.when_held, set())
                self._ended_stack.update(commands)

    def _reset_all_stacks(self) -> None:
        self._all_stack = set()
        self._actions_stack = {}
        self._incoming_stack = set()
        self._scheduled_stack = set()
        self._ended_stack = set()
        self._cancel_stack = set()
        self._subsystem_stack = set()

    def _schedule_default_commands(self) -> None:  # noqa: WPS231
        for subsystem in self._subsystem_stack:
            if subsystem.default_command and not subsystem.current_command:
                for command in self._incoming_stack.union(self._scheduled_stack):
                    if subsystem in command.requirements:
                        break
                else:
                    self._incoming_stack.add(subsystem.default_command)

    def _update_stack(self) -> None:
        # Remove interrupted and ended commands
        self._scheduled_stack.difference_update(self._ended_stack)
        self._scheduled_stack.difference_update(self._cancel_stack)

        # Move incoming commands to scheduled stack
        self._scheduled_stack.update(self._incoming_stack)
        self._all_stack.update(self._scheduled_stack)

        # Reset incoming, ended, and interrupted stacks
        self._incoming_stack.clear()
        self._ended_stack.clear()
        self._cancel_stack.clear()



from contextlib import suppress
from typing import Optional

from command_based_framework._common import ContextManagerMixin
from command_based_framework.scheduler import Scheduler

with suppress(ImportError):
    from command_based_framework.commands import Command


class Subsystem(ContextManagerMixin):
    """Breaks out complex robot components into methods and attributes.

    Subsystems define how something is performed; i.e. reading a sensor.
    Subsystems are also used by the scheduler to ensure two commands are
    not using the same resources simultaneously.
    """

    # The name of the subsystem
    _name: str

    # The command that is currently using this subsystem
    _current_command: Optional["Command"]

    # The command that will run whenever this subsystem is not being
    # used
    _default_command: Optional["Command"]

    def __init__(self, name: Optional[str] = None) -> None:
        """Creates a new `Subsystem` instance.

        When created, the subsystem will automatically register itself
        with the scheduler.

        :param name: The name of the subsystem. If not provided, the
            class name is used instead.
        :type name: str, optional
        """
        super().__init__()
        self._name = name or self.__class__.__name__
        self._current_command = None
        self._default_command = None

        # Register this subsystem in the scheduler's stack
        Scheduler.instance.register_subsystem(self)  # type: ignore

    @property
    def current_command(self) -> Optional["Command"]:
        """The command that is currently using this subsystem.

        This property is controlled by the scheduler and should not be
        modified directly.

        If no default command is set and no command is currently using
        this subsystem, then this property will be `None`.
        """
        return self._current_command

    @current_command.setter
    def current_command(self, command: Optional["Command"]) -> None:
        self._current_command = command

    @property
    def default_command(self) -> Optional["Command"]:
        """The command to run when no other command is active.

        If not specified, then this subsystem will remain idle.

        Default commands **must** require the subsystems they are
        assigned to. A :py:exc:`ValueError` will be raised otherwise.
        """
        return self._default_command

    @default_command.setter
    def default_command(self, command: Optional["Command"]) -> None:
        # Ensure the command requires this subsystem
        if command and self not in command.requirements:
            raise ValueError(
                (
                    "{name} must have {name2} as a requirement "
                    "before being assigned as a default"
                ).format(
                    name=command.name,
                    name2=self.name,
                ),
            )
        self._default_command = command

    @property
    def name(self) -> str:
        """The name of the subsystem.

        This is a read-only property.

        If one was not provided at the creation of the command, the
        class name is used instead.
        """
        return self._name

    def periodic(self) -> None:
        """Periodically called when the subsystem is required by a scheduled :py:class:`~command_based_framework.commands.Command`.

        Override this behavior to always execute by calling
        :py:meth:`~command_based_framework.scheduler.Scheduler.register_subsystem`.
        """  # noqa: E501



from abc import ABC, abstractmethod
from enum import Enum, auto

from command_based_framework._common import ContextManagerMixin
from command_based_framework.commands import Command
from command_based_framework.scheduler import Scheduler


class Condition(Enum):
    """Enums representing different action conditions."""

    cancel_when_activated = auto()
    toggle_when_activated = auto()
    when_activated = auto()
    when_deactivated = auto()
    when_held = auto()


class Action(ABC, ContextManagerMixin):  # noqa: WPS214
    """Schedules :py:class:`~command_based_framework.commands.Command` based on a condition being met.

    Actions determine when commands are scheduled/executed. To do this,
    the scheduler periodically runs the :py:meth:`~command_based_framework.actions.Action.poll`
    method. Any arbitrary condition, or multiple conditions, can be
    implemented in :py:meth:`~command_based_framework.actions.Action.poll`.

    To setup when commands are scheduled, bind them using any of the
    `when` methods. Attempts to bind a command multiple times in the
    same action will result in the previous binding being replaced by
    the new one.

    A command can be bound to a single or multiple actions.
    """  # noqa: E501

    # Flag to indicate the state of the action the last time it was
    # checked
    _last_state: bool

    def __init__(self) -> None:
        """Creates a new `Action` instance."""
        super().__init__()
        self._last_state = False

    @property
    def last_state(self) -> bool:
        """The state of the action the last time it was checked."""
        return self._last_state

    @last_state.setter
    def last_state(self, state: bool) -> None:
        self._last_state = state

    @abstractmethod
    def poll(self) -> bool:
        """Check if the condition to activate commands are met.

        Only return `True` when all conditions are met for this action
        to activate and schedule bound commands.
        """
        return False  # pragma: no cover

    def cancel_when_activated(self, command: Command) -> None:
        """Cancel `command` when this action is activated."""
        Scheduler.instance.bind_command(  # type: ignore
            self,
            command,
            Condition.cancel_when_activated,
        )

    def toggle_when_activated(self, command: Command) -> None:
        """Toggle scheduling `command` when this action is activated.

        For example, a button is pressed for the first time and a
        command runs. The same button is pressed again, but the command
        exits. The cycle repeats when the button is pressed for a third
        time.
        """
        Scheduler.instance.bind_command(  # type: ignore
            self,
            command,
            Condition.toggle_when_activated,
        )

    def when_activated(self, command: Command) -> None:
        """Schedule `command` when this action is activated."""
        Scheduler.instance.bind_command(  # type: ignore
            self,
            command,
            Condition.when_activated,
        )

    def when_deactivated(self, command: Command) -> None:
        """Schedule `command` when this action is deactivated."""
        Scheduler.instance.bind_command(  # type: ignore
            self,
            command,
            Condition.when_deactivated,
        )

    def when_held(self, command: Command) -> None:
        """Schedule `command` when this action is perpetually activated."""
        Scheduler.instance.bind_command(  # type: ignore
            self,
            command,
            Condition.when_held,
        )



from types import TracebackType
from typing import Optional, Type


class ContextManagerMixin(object):
    """Mixin providing context manager support.

    Also contains a :py:meth:`~command_based_framework._command.ContextManagerMixin.handle_exception`
    which is called by the scheduler whenever the parent child instance
    raises an error. Use this method to process the exception.
    """  # noqa: E501

    def __enter__(self) -> "ContextManagerMixin":
        """Called when the command enters a context manager."""
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]] = None,
        exc: Optional[BaseException] = None,
        traceback: Optional[TracebackType] = None,
    ) -> bool:
        """Called when the command exits a context manager."""
        # Ignore non-errors
        if not exc_type or not exc or not traceback:
            return True

        self.handle_exception(exc_type, exc, traceback)
        return True

    def handle_exception(
        self,
        exc_type: Type[BaseException],
        exc: BaseException,
        traceback: TracebackType,
    ) -> bool:
        """Called when any method raises an error.

        If a command, the scheduler uses the output of this method to
        determine whether the command should be immediately interrupted.

        :param exc_type: The type of exception raised.
        :type exc_type: :py:class:`Type`
        :param exc: The exception raised.
        :type exc: :py:class:`Exception`
        :param traceback: The frame traceback of the error.
        :type traceback: :py:class:`Traceback`

        :return: `True` to indicate the error is handled. All other
            returns to the scheduler will be interpreted as the command
            needing to be immediately interrupted, if the child class is
            a command.
        :rtype: bool
        """  # noqa: DAR202
        return False  # pragma: no cover
